"""File related hooks

:organization: Logilab
:copyright: 2003-2018 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
:contact: http://www.logilab.fr/ -- mailto:contact@logilab.fr
"""

import os

from cubicweb.server import hook
from cubicweb.sobjects.notification import ContentAddedView
from cubicweb.predicates import is_instance

from cubicweb_file.entities import UnResizeable


class UpdateFileHook(hook.Hook):
    """a file has been updated, check data_format/data_encoding consistency
    """
    __regid__ = 'updatefilehook'
    __select__ = hook.Hook.__select__ & is_instance('File')
    events = ('before_add_entity', 'before_update_entity',)
    order = -1  # should be run before other hooks
    category = 'hash'

    def __call__(self):
        edited = self.entity.cw_edited
        if 'data' in edited:
            self.entity.set_format_and_encoding()
            maxsize = self._cw.vreg.config['image-max-size']
            if maxsize and self.entity.data_format.startswith('image/'):
                iimage = self.entity.cw_adapt_to('IImage')
                try:
                    edited['data'] = iimage.resize(maxsize)
                except UnResizeable:
                    # if the resize fails for some reason, do nothing
                    # (original image will be stored)
                    pass

            if self._cw.vreg.config['compute-hash']:
                data = edited['data']
                if data is not None:
                    hashdata = self.entity.compute_hash(data.getvalue())
                    edited['data_hash'] = hashdata

            # thumbnail cache invalidation
            if 'update' in self.event and 'data' in edited:
                thumb = self.entity.cw_adapt_to('IThumbnail')
                if not thumb:
                    return
                thumbpath = thumb.thumbnail_path()
                if thumbpath:
                    try:
                        os.unlink(thumbpath)
                    except Exception as exc:
                        self.warning(
                            'could not invalidate thumbnail file `%s` '
                            '(cause: %s)',
                            thumbpath, exc)


class FileAddedView(ContentAddedView):
    """get notified from new files"""
    __select__ = is_instance('File')
    content_attr = 'description'



import hashlib

options = (
    ('image-max-size',
     {'type': 'string',
      'default': None,
      'help': 'all images will be resized to this max size if set',
      'group': 'file',
      'level': 2,
      }),
    ('image-thumb-size',
     {'type': 'string',
      'default': '75x75',
      'help': 'thumbnail size of your images',
      'group': 'file',
      'level': 2,
      }),
    ('thumbnail-cache-directory',
     {'type': 'string',
      'default': None,
      'help': ('Cache directory for thumbnails (if unset, defaults '
               'to INSTANCEDIR/cache/file/thumbnails) '
               'if set, must be an absolute path).'),
      'level': 2,
      'group': 'file'
      }),
    ('compute-hash',
     {'type': 'yn',
      'default': False,
      'help': 'compute hash digest of your files',
      'group': 'file',
      'level': 2,
      }),
    ('hash-algorithm',
     {'type': 'choice',
      'default': 'sha256',
      'choices': list(hashlib.algorithms_available),
      'help': ('hash algorithm used to compute hash of your files (%s)' %
               ', '.join(sorted(hashlib.algorithms_available))),
      'group': 'file',
      'level': 2,
      }),
    )



from cubicweb.devtools.testlib import AutomaticWebTest


class AutomaticWebTest(AutomaticWebTest):

    def to_test_etypes(self):
        return set(('File',))

    def list_startup_views(self):
        return ()


if __name__ == '__main__':
    from unittest import main
    main()



from cubicweb.web.views.urlrewrite import SimpleReqRewriter, rgx


class FileReqRewriter(SimpleReqRewriter):
    rules = [
         (rgx(r'/file/(\d+)/raw(/.*)?'),
          dict(rql=r'Any X WHERE X eid \1', vid='download')),
         (rgx(r'/file/(\d+)/thumb(/.*)?'),
          dict(rql=r'Any X WHERE X eid \1', vid='thumbnail'))
    ]



"""cubicweb-ctl plugin providing the fsimport command

:organization: Logilab
:copyright: 2010-2018 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
:contact: http://www.logilab.fr/ -- mailto:contact@logilab.fr
"""

import hashlib

from cubicweb.cwctl import CWCTL
from cubicweb.toolsutils import Command


class FSImportCommand(Command):
    """Import content of a directory on the file system as File entities.
    The instance must use the `file` cube.

    <instance id>
      identifier of the instance where directory's content has to be imported.

    <fs directory>
      directory to import (recursivly)
    """
    name = 'fsimport'
    min_args = 2
    arguments = '<instance id> <fs file or directory>...'
    options = (
        ("map-folders",
         {'short': 'F', 'action': 'store_true',
          'default': False,
          'help': ('map file-system directories as Folder entities '
                   '(requires the `folder` cube).'),
          }),
        ("filed-under",
         {'short': 'u', 'type': 'int',
          'help': 'put imported file into the folder entity of the given eid.',
          }),
        )

    def run(self, args):
        """run the command with its specific arguments"""
        from cubicweb.server.serverconfig import ServerConfiguration
        from cubicweb.server.serverctl import repo_cnx
        from cubicweb_file.fsimport import fsimport
        appid = args.pop(0)
        config = ServerConfiguration.config_for(appid)
        repo, cnx = repo_cnx(config)
        repo.hm.call_hooks('server_maintenance', repo=repo)
        with cnx:
            fsimport(cnx, args, parenteid=self.config.filed_under,
                     mapfolders=self.config.map_folders,
                     bfss='data' in repo.system_source._storages.get(
                         'File', ()))
            cnx.commit()


class FileRefreshHashCommand(Command):
    """Import content of a directory on the file system as File entities.
    The instance must use the `file` cube.

    <instance id>
      identifier of the instance where directory's content has to be imported.

    """
    name = 'refresh-hash'
    min_args = 1
    arguments = '<instance id>...'
    options = (
        ("hash-algorithm",
         {'default': None,
          'type': 'choice',
          'choices': list(hashlib.algorithms_available),
          'help': ('hash algorithm to use (instead of the one configured '
                   'for the instance)'),
          }),
        ("force",
         {'short': 'f', 'action': 'store_true',
          'default': False,
          'help': ("Force the computation of hash if the instance "
                   "'compute-hash' option if False."),
          }),
        ("subclasses",
         {'action': 'store_true',
          'default': False,
          'help': "Also refresh hash for entity types that inherit from File",
          }),
        )

    def run(self, args):
        """run the command with its specific arguments"""
        from cubicweb.server.serverconfig import ServerConfiguration
        from cubicweb.server.serverctl import repo_cnx
        appid = args.pop(0)
        config = ServerConfiguration.config_for(appid)

        repo, cnx = repo_cnx(config)
        if repo.vreg.config['compute-hash'] or self.config.force:
            repo.hm.call_hooks('server_maintenance', repo=repo)
            alg = self.config.hash_algorithm

            with cnx:
                with cnx.deny_all_hooks_but():
                    etypes = ['File']
                    if self.config.subclasses:
                        etypes.extend(cnx.repo.schema['File'].specialized_by(
                            recursive=True))
                    print('ETYPES=', etypes)
                    for etype in etypes:
                        for e in cnx.execute(
                                'Any X WHERE X is %s' % etype).entities():
                            print('update hash for', e)
                            e.cw_set(data_hash=e.compute_hash(alg=alg))
                    cnx.commit()


CWCTL.register(FSImportCommand)



from yams.buildobjs import RelationDefinition
from cubicweb_folder.schema import Folder
from cubicweb_file.schema import File

# XXX dirty hack to be quite sure tests won't fail due to path being >
# 64 chars
Folder.get_relation('name').constraints[0].max = 256


class filed_under(RelationDefinition):
    subject = 'File'
    object = 'Folder'


class MyFile(File):
    __specializes_schema__ = True



# -*- coding: utf-8 -*-

from os.path import join, dirname
from PIL.Image import open as pilopen

from cubicweb import Binary, Unauthorized
from cubicweb.devtools.testlib import CubicWebTC


class FileTC(CubicWebTC):

    def test_set_mime_and_encoding(self):
        with self.admin_access.client_cnx() as cnx:
            fobj = cnx.create_entity(
                'File', data_name=u"foo.txt", data=Binary(b"xxx"))
            self.assertEqual(fobj.data_format, u'text/plain')
            self.assertEqual(fobj.data_encoding, cnx.encoding)

    def test_set_mime_and_encoding_gz_file(self):
        with self.admin_access.client_cnx() as cnx:
            fobj = cnx.create_entity(
                'File', data_name=u"foo.txt.gz", data=Binary(b"xxx"))
            self.assertEqual(fobj.data_format, u'text/plain')
            self.assertEqual(fobj.data_encoding, u'gzip')
            fobj = cnx.create_entity(
                'File', data_name=u"foo.txt.gz", data=Binary(b"xxx"),
                data_format='application/gzip')
            self.assertEqual(fobj.data_format, u'text/plain')
            self.assertEqual(fobj.data_encoding, u'gzip')
            fobj = cnx.create_entity(
                'File', data_name=u"foo.gz", data=Binary(b"xxx"))
            self.assertEqual(fobj.data_format, u'application/gzip')
            self.assertEqual(fobj.data_encoding, None)

    def test_set_mime_and_encoding_bz2_file(self):
        with self.admin_access.client_cnx() as cnx:
            fobj = cnx.create_entity(
                'File', data_name=u"foo.txt.bz2", data=Binary(b"xxx"))
            self.assertEqual(fobj.data_format, u'text/plain')
            self.assertEqual(fobj.data_encoding, u'bzip2')
            fobj = cnx.create_entity(
                'File', data_name=u"foo.txt.bz2", data=Binary(b"xxx"),
                data_format='application/bzip2')
            self.assertEqual(fobj.data_format, u'text/plain')
            self.assertEqual(fobj.data_encoding, u'bzip2')
            fobj = cnx.create_entity(
                'File', data_name=u"foo.bz2", data=Binary(b"xxx"))
            self.assertEqual(fobj.data_format, u'application/bzip2')
            self.assertEqual(fobj.data_encoding, None)

    def test_set_mime_and_encoding_unknwon_ext(self):
        with self.admin_access.client_cnx() as cnx:
            fobj = cnx.create_entity(
                'File', data_name=u"foo.123", data=Binary(b"xxx"))
            self.assertEqual(fobj.data_format, u'application/octet-stream')
            self.assertEqual(fobj.data_encoding, None)


class ImageTC(CubicWebTC):

    @property
    def data(self):
        with open(join(dirname(__file__), 'data', '20x20.gif'), 'rb') as fobj:
            return fobj.read()

    def test_resize_image(self):
        with self.admin_access.client_cnx() as cnx:
            # check no resize
            img = cnx.create_entity(
                'File', data=Binary(self.data), data_name=u'20x20.gif')
            self.assertEqual(img.data_format, u'image/gif')
            self.assertEqual(img.data.getvalue(), self.data)
            # check thumb
            self.set_option('image-thumb-size', '5x5')
            pilthumb = pilopen(
                img.cw_adapt_to('IImage').thumbnail(shadow=False))
            self.assertEqual(pilthumb.size, (5, 5))
            self.assertEqual('PNG', pilthumb.format)
            # check resize 10x5
            self.set_option('image-max-size', '10x5')
            img = cnx.create_entity(
                'File', data=Binary(self.data), data_name=u'20x20.gif')
            self.assertEqual(img.data_format, u'image/gif')
            pilimg = pilopen(img.data)
            self.assertEqual(pilimg.size, (5, 5))
            # also on update
            img.cw_set(data=Binary(self.data))
            img.cw_clear_all_caches()
            pilimg = pilopen(img.data)
            self.assertEqual(pilimg.size, (5, 5))
            # test image smaller than max size
            self.set_option('image-max-size', '40x40')
            img.cw_set(data=Binary(self.data))
            pilimg = pilopen(img.data)
            self.assertEqual(pilimg.size, (20, 20))


class HashTC(CubicWebTC):

    def test_init_hash(self):
        with self.admin_access.client_cnx() as cnx:
            cnx.vreg.config['compute-hash'] = 0
            fobj = cnx.create_entity(
                'File', data_name=u"foo.txt", data=Binary(b"xxx"))
            self.assertEqual(None, fobj.data_hash)

            cnx.vreg.config['compute-hash'] = 1
            cnx.vreg.config['hash-algorithm'] = 'sha256'
            fobj = cnx.create_entity(
                'File', data_name=u"foo.txt", data=Binary(b"xxx"))
            self.assertEqual(
                '{sha256}cd2eb0837c9b4c962c22d2ff8b5441b7b45805887f051d39bf133b583baf6860',  # noqa
                fobj.data_hash)

    def test_modify_data(self):
        with self.admin_access.client_cnx() as cnx:
            cnx.vreg.config['compute-hash'] = 1
            cnx.vreg.config['hash-algorithm'] = 'sha256'
            fobj = cnx.create_entity(
                'File', data_name=u"foo.txt", data=Binary(b"xxx"))
            fobj.cw_set(data=Binary(b'yyy'))
            self.assertEqual(
                '{sha256}f2afd1cacb5441a5e65a7a460a5f9898b7b98b08aa6323a2e53c8b9a9686cd86',  # noqa
                fobj.data_hash)

    def test_manual_set_hash_forbidden(self):
        with self.admin_access.client_cnx() as cnx:
            cnx.vreg.config['compute-hash'] = 0
            with self.assertRaises(Unauthorized):
                cnx.create_entity(
                    'File', data_name=u"foo.txt", data=Binary(b"xxx"),
                    data_hash=u'0'*40)
                cnx.commit()

    def test_hash_algos(self):
        with self.admin_access.client_cnx() as cnx:
            cnx.vreg.config['compute-hash'] = 1
            cnx.vreg.config['hash-algorithm'] = 'sha256'
            fobj = cnx.create_entity(
                'File', data_name=u"foo.txt", data=Binary(b"xxx")).eid
            cnx.commit()
        with self.admin_access.client_cnx() as cnx:
            fobj = cnx.find('File', eid=fobj).one()

            cnx.vreg.config['hash-algorithm'] = 'md5'
            fobj.cw_set(data=Binary(b'md5'))
            fobj.cw_clear_all_caches()
            self.assertEqual(
                '{md5}1bc29b36f623ba82aaf6724fd3b16718',
                fobj.data_hash)

            cnx.vreg.config['hash-algorithm'] = 'sha1'
            fobj.cw_set(data=Binary(b'sha1'))
            fobj.cw_clear_all_caches()
            self.assertEqual(
                '{sha1}415ab40ae9b7cc4e66d6769cb2c08106e8293b48',
                fobj.data_hash)

            cnx.vreg.config['hash-algorithm'] = 'sha512'
            fobj.cw_set(data=Binary(b'sha512'))
            fobj.cw_clear_all_caches()
            self.assertEqual(
                '{sha512}1f9720f871674c18e5fecff61d92c1355cd4bfac25699fb7ddfe7717c9669b4d085193982402156122dfaa706885fd64741704649795c65b2a5bdec40347e28a',  # noqa
                fobj.data_hash)


if __name__ == '__main__':
    from unittest import main
    main()




from cubicweb import Binary
from cubicweb.devtools import testlib, ApptestConfiguration
from cubicweb_file.ccplugin import FileRefreshHashCommand
from cubicweb.server.serverconfig import ServerConfiguration


class FileRefreshCommandTC(testlib.CubicWebTC):

    def setUp(self):
        super(FileRefreshCommandTC, self).setUp()
        self.orig_config_for = ServerConfiguration.config_for

        def config_for(appid):
            return ApptestConfiguration(appid, __file__)

        ServerConfiguration.config_for = staticmethod(config_for)

    def tearDown(self):
        ServerConfiguration.config_for = self.orig_config_for
        super(FileRefreshCommandTC, self).tearDown()

    def test_refresh(self):
        with self.admin_access.repo_cnx() as cnx:
            cnx.vreg.config['compute-hash'] = 0
            for i in range(10):
                fobj = cnx.create_entity(
                    'File', data_name=u"foo%d.txt" % i,
                    data_format=u'text/plain',
                    data=Binary(b"xxx"))
                self.assertEqual(None, fobj.data_hash)
            for i in range(10):
                fobj = cnx.create_entity(
                    'MyFile', data_name=u"foo%s.png" % i,
                    data_format=u'image/png',
                    data=Binary(b"xxx"))
                self.assertEqual(None, fobj.data_hash)
            cnx.commit()
        FileRefreshHashCommand(None).run([self.appid])
        with self.admin_access.repo_cnx() as cnx:
            self.assertFalse(cnx.execute('Any X WHERE X is_instance_of File, '
                                         'NOT X data_hash NULL'))

        cmd = FileRefreshHashCommand(None)
        cmd.config.force = True
        cmd.run([self.appid])
        with self.admin_access.repo_cnx() as cnx:
            self.assertEqual(
                10,
                len(cnx.execute('Any X WHERE X is_instance_of File, '
                                'X data_hash LIKE "{sha256}%"')))

        cmd.config.subclasses = True
        cmd.run([self.appid])
        with self.admin_access.repo_cnx() as cnx:
            self.assertEqual(
                20,
                len(cnx.execute('Any X WHERE X is_instance_of File, '
                                'X data_hash LIKE "{sha256}%"')))


if __name__ == '__main__':
    from unittest import main
    main()



"""hash functions for the cubicbweb-file cube

:organization: Logilab
:copyright: 2018 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
:contact: http://www.logilab.fr/ -- mailto:contact@logilab.fr
"""

import re
import hashlib
from six import text_type


def compute_hash(value, alg):
    hashvalue = compute_row_hash(value, alg)
    if hashvalue:
        return '{%s}%s' % (alg, hashvalue)


def compute_row_hash(value, alg):
    if value is not None:
        hasher = hashlib.new(alg, value)
        return text_type(hasher.hexdigest())


def check_hash(hash_value, value):
    m = re.match('{(?P<alg>[A-Za-z0-9]+)}(?P<hash>.+)', hash_value)
    if m:
        alg = m.group('alg')
        hashhex = m.group('hash')
    else:  # bw compat, used to be sha1
        alg = 'sha1'
        hashhex = hash_value
    comphash = compute_row_hash(value, alg=alg)
    return comphash == hashhex



"""file/image gallery view

:organization: Logilab
:copyright: 2006-2013 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
:contact: http://www.logilab.fr/ -- mailto:contact@logilab.fr
"""

__docformat__ = "restructuredtext en"

from logilab.mtconverter import xml_escape

from cubicweb.view import EntityView
from cubicweb.predicates import is_instance, has_mimetype, adaptable
from cubicweb.web.views import ajaxcontroller
from cubicweb.web import NotFound


class GalleryView(EntityView):
    __regid__ = 'gallery'
    __select__ = is_instance('File')

    currently_displayed = None

    def call(self):
        self._cw.add_js(('cubes.file.js', 'cubicweb.ajax.js'))
        self._cw.add_css('cubes.file.css')
        try:
            eid = int(self._cw.form.get('selected', self.cw_rset[0][0]))
        except ValueError:
            raise NotFound()
        self.currently_displayed = eid
        rset = self._cw.execute('Any X where X eid %(x)s', {'x': eid})
        if not rset:
            raise NotFound()
        for i in range(self.cw_rset.rowcount):
            self.cell_call(row=i, col=0)
        self.w(u'<div id="imageholder">')
        self.wview('primary', rset, row=0, col=0, initargs={'is_primary': False})
        self.w(u'</div>')
        self.w(u'<div class="imagegallery">')
        self.w(u'</div>')

    def cell_call(self, row, col):
        entity = self.cw_rset.complete_entity(row, col)
        ithumb = entity.cw_adapt_to('IThumbnail')
        if ithumb is not None:
            icon = xml_escape(ithumb.thumbnail_url())
        else:
            icon = self._cw.uiprops['FILE_ICON']
        title = xml_escape(entity.dc_title())
        # .currently_displayed is None if we went directly to cell_call,
        # eg view called only for that entity, in which case we want
        # it to be considered as selected.
        if self.currently_displayed is None or entity.eid == self.currently_displayed:
            self.w(u'<a href="javascript:displayImg(%(eid)s)" title="%(title)s"><img id="img%(eid)s" '
                   u'class="selectedimg" alt="%(title)s" src="%(icon)s"/></a>'
                   % {'eid': entity.eid, 'title': title, 'icon': icon})
        else:
            self.w(u'<a href="javascript:displayImg(%(eid)s)" title="%(title)s"><img id="img%(eid)s" '
                   'alt="%(title)s" src="%(icon)s"/></a>'
                   % {'eid': entity.eid, 'title': title, 'icon': icon})


@ajaxcontroller.ajaxfunc(output_type='xhtml')
def get_image(self, eid):
    return self._cw.view('primary', self._cw.eid_rset(eid), row=0, col=0,
                         initargs={'is_primary': False})


class AlbumView(EntityView):
    __regid__ = 'album'
    __select__ = adaptable('IImage')  # good enough for IThumbnail

    def call(self, nbcol=5):
        self._cw.add_css('cubes.file.css')
        lines = [[]]
        for idx in range(self.cw_rset.rowcount):
            if len(lines[-1]) == nbcol:
                lines.append([])
            lines[-1].append(self._make_cell(idx, 0))
        while len(lines[-1]) != nbcol:
            lines[-1].append(u'&#160;')
        self.w(u'<table class="album">')
        for line in lines:
            self.w(u'<tr>')
            self.w(u''.join(u'<td>%s</td>' % cell for cell in line))
            self.w(u'</tr>')
        self.w(u'</table>')

    def _make_cell(self, row, col):
        entity = self.cw_rset.complete_entity(row, col)
        icon = xml_escape(entity.cw_adapt_to('IThumbnail').thumbnail_url())
        title = xml_escape(entity.dc_title())
        return (u'<a href="%(url)s" title="%(title)s"><img alt="%(title)s" src="%(icon)s"/></a>'
                % {'url': xml_escape(entity.absolute_url()), 'title': title, 'icon': icon})

    def cell_call(self, row, col):
        self.w(self._make_cell(row, col))


class ImageAdaptedView(AlbumView):
    __regid__ = 'sameetypelist'
    __select__ = has_mimetype('image/') & adaptable('IImage')



from __future__ import print_function

import sys
from os import listdir, path as osp

from six import text_type, binary_type, PY3

from cubicweb import Binary


def fsimport(cw, fspaths, parenteid=None, mapfolders=True, bfss=False,
             fsencoding=sys.stdout.encoding or 'utf-8', quiet=False):
    assert bfss, 'not implemented'
    if bfss:
        cw.transaction_data['fs_importing'] = True
    if parenteid is not None:
        parent = cw.execute('Any X WHERE X eid %(x)s',
                            {'x': parenteid}).get_entity(0, 0)
    else:
        parent = None
    # XXX what if not bfss?
    alreadyimported = set(x.getvalue() for x, in cw.execute(
        'Any fspath(D) WHERE X data D, X is File'))
    for fspath in fspaths:
        fspath = osp.abspath(osp.normpath(fspath))
        if osp.isdir(fspath):
            import_directory(cw, fspath, parent, mapfolders, bfss, fsencoding,
                             quiet, _alreadyimported=alreadyimported)
        else:
            import_file(cw, fspath, parent, bfss, fsencoding, quiet,
                        _alreadyimported=alreadyimported)


def import_directory(cw, directory, parent=None, mapfolders=True, bfss=False,
                     fsencoding=sys.stdout.encoding or 'utf-8', quiet=False,
                     _indent='', _alreadyimported=None):
    if _alreadyimported is None:
        _alreadyimported = set()
    # getting directory content will check it's validity, do it before creating
    # any Folder entity
    dirfiles = listdir(directory)
    # XXX what if not bfss
    if mapfolders:
        parent = folder_entity(cw, directory, parent=parent,
                               fsencoding=fsencoding)
    if not quiet:
        print(_indent + '** importing directory', directory)
    for fname in dirfiles:
        fspath = osp.join(directory, fname)
        if osp.isdir(fspath):
            import_directory(cw, fspath, parent, mapfolders, bfss, fsencoding,
                             quiet, _indent=_indent + '  ',
                             _alreadyimported=_alreadyimported)
        else:
            import_file(cw, fspath, parent, bfss, fsencoding, quiet,
                        _indent=_indent + '  ',
                        _alreadyimported=_alreadyimported)


def import_file(cw, fspath, parent=None, bfss=False,
                fsencoding=sys.stdout.encoding or 'utf-8', quiet=False,
                _indent='', _alreadyimported=None):
    fname = osp.basename(fspath)
    if isinstance(fname, binary_type):
        fname = fname.decode(fsencoding)
    if PY3 and isinstance(fspath, text_type):
        fspath = fspath.encode(fsencoding)
    if _alreadyimported is None:
        _alreadyimported = set()
    if fspath in _alreadyimported:
        if not quiet:
            print(_indent + '  skipping already imported', fname)
        return
    if not quiet:
        print(_indent + '  importing', fname)
    if bfss:
        efile = cw.create_entity('File', data_name=fname,
                                 data=Binary(fspath))
    else:
        with open(fspath, 'rb') as stream:
            efile = cw.create_entity('File', data_name=fname,
                                     data=Binary(stream.read()))
    _alreadyimported.add(fspath)
    if parent is not None:
        efile.cw_set(filed_under=parent)


def folder_entity(cw, directory, parent=None, fsencoding='utf-8'):
    if isinstance(directory, binary_type):
        directory = directory.decode(fsencoding)
    if parent is None:
        rset = cw.execute('Folder X WHERE X name %(name)s',
                          {'name': directory})
    else:
        rset = cw.execute('Folder X WHERE X name %(name)s, '
                          'X filed_under P, P eid %(p)s',
                          {'name': osp.dirname(directory), 'p': parent.eid})
    if rset:
        return rset.get_entity(0, 0)
    # create the folder
    if parent is None:
        return cw.create_entity('Folder', name=directory)
    return cw.create_entity('Folder', name=directory, filed_under=parent)



from os.path import dirname, join, abspath

from six import PY3

from cubicweb.devtools.testlib import CubicWebTC
from cubicweb.server.sources import storages

from cubicweb_file.fsimport import fsimport

HERE = dirname(__file__)

STORAGE = storages.BytesFileSystemStorage('whatever')


class importDirectoryTC(CubicWebTC):

    def setup_database(self):
        storages.set_attribute_storage(self.repo, 'File', 'data', STORAGE)

    def tearDown(self):
        storages.unset_attribute_storage(self.repo, 'File', 'data')
        super(importDirectoryTC, self).tearDown()

    def test_folder_bfss(self):
        with self.admin_access.repo_cnx() as cnx:
            fsimport(cnx, [join(HERE, 'data', 'toimport')],
                     bfss=True, quiet=True)
            self.assertEqual(
                cnx.execute('Any COUNT(F) WHERE F is Folder')[0][0], 1)
            self.assertEqual(
                cnx.execute('Any COUNT(F) WHERE F is File')[0][0], 2)
            fpath = cnx.execute(
                'Any fspath(D) WHERE F is File, F data D, '
                'F data_name "coucou.txt"')[0][0].getvalue()
            if PY3:
                fpath = fpath.decode()
            self.assertEqual(fpath, join(
                abspath(HERE), 'data', 'toimport', 'coucou.txt'))
            f = cnx.execute(
                'Any F WHERE F is File, '
                'F data_name "coucou.txt"').get_entity(0, 0)
            self.assertEqual(f.data.getvalue(), b'bijour\n')
            self.assertEqual(len(f.filed_under), 1)
            self.assertEqual(f.filed_under[0].name, join(
                abspath(HERE), 'data', 'toimport'))
            # test reimport
            fsimport(cnx, [join(HERE, 'data', 'toimport')],
                     bfss=True, quiet=True)
            self.assertEqual(
                cnx.execute('Any COUNT(F) WHERE F is Folder')[0][0], 1)
            self.assertEqual(
                cnx.execute('Any COUNT(F) WHERE F is File')[0][0], 2)

    def test_file_bfss(self):
        with self.admin_access.repo_cnx() as cnx:
            f = cnx.create_entity('Folder', name=u'toto')
            fsimport(cnx, [join(HERE, 'data', 'toimport', 'IMG_8033.jpg'),
                           join(HERE, 'data', 'toimport', 'coucou.txt')],
                     parenteid=f.eid, bfss=True, quiet=True)
            self.assertEqual(
                cnx.execute('Any COUNT(F) WHERE F is Folder')[0][0], 1)
            self.assertEqual(
                cnx.execute('Any COUNT(F) WHERE F is File')[0][0], 2)
            fpath = cnx.execute(
                'Any fspath(D) WHERE F is File, F data D, '
                'F data_name "coucou.txt"')[0][0].getvalue()

            if PY3:
                fpath = fpath.decode()
            self.assertEqual(fpath, join(
                abspath(HERE), 'data', 'toimport', 'coucou.txt'))
            f = cnx.execute(
                'Any F WHERE F is File, '
                'F data_name "coucou.txt"').get_entity(0, 0)
            self.assertEqual(f.filed_under[0].name, 'toto')

    def test_unexistant(self):
        # test import of an unexistant directory
        with self.admin_access.repo_cnx() as cnx:
            self.assertRaises(
                IOError, fsimport, cnx, [join(HERE, 'pouet')],
                bfss=True, quiet=True)
            self.assertEqual(
                cnx.execute('Any COUNT(F) WHERE F is Folder')[0][0], 0)


if __name__ == '__main__':
    import unittest
    unittest.main()



# pylint: disable-msg=E0611,F0401

from cubicweb import _

from yams.buildobjs import EntityType, String, Bytes, RichString


class File(EntityType):
    """a downloadable file which may contains binary data"""
    title = String(fulltextindexed=True, maxsize=256)
    data = Bytes(required=True, description=_('file to upload'))
    data_format = String(
        required=True, maxsize=128,
        description=_('MIME type of the file. Should be dynamically set '
                      'at upload time.'))
    data_encoding = String(
        maxsize=32,
        description=_('encoding of the file when it applies (e.g. text). '
                      'Should be dynamically set at upload time.'))
    data_name = String(
        required=True, fulltextindexed=True,
        description=_('name of the file. Should be dynamically set '
                      'at upload time.'))
    data_hash = String(
        maxsize=256,  # max len of currently available hash alg + prefix is 140
        description=_('hash of the file. May be set at upload time.'),
        __permissions__={'read': ('managers', 'users', 'guests'),
                         'add': (),
                         'update': ()})
    description = RichString(fulltextindexed=True, internationalizable=True,
                             default_format='text/rest')



# coding: utf-8

from os.path import join, dirname, isfile, exists
import shutil

from cubicweb.devtools.testlib import CubicWebTC

from cubicweb import NoSelectableObject, Binary, Unauthorized
from cubicweb.web import NotFound

from cubicweb_file.entities import thumb_cache_dir


class FileTC(CubicWebTC):
    icon = 'text.ico'
    mime_type = u"text/plain"

    def setup_database(self):
        with self.admin_access.repo_cnx() as cnx:
            create = cnx.create_entity
            self.fobj = create(
                'File', data_name=u"foo.pdf",
                data=Binary(b"xxx"),
                data_format=self.mime_type).eid
            self.ufobj = create(
                'File', data_name=u"Bâbâr.pdf",
                data=Binary(b"yyy"),
                data_format=self.mime_type).eid
            cnx.commit()

    def test_idownloadable(self):
        with self.admin_access.client_cnx() as cnx:
            fobj = cnx.entity_from_eid(self.fobj)
            idownloadable = fobj.cw_adapt_to('IDownloadable')
            self.assertEqual(idownloadable.download_data(), b'xxx')
            self.assertEqual(idownloadable.download_url(),
                             u'http://testing.fr/cubicweb/%s/%s/raw/%s' % (
                fobj.__regid__.lower(), self.fobj, fobj.data_name))
            self.assertEqual(
                idownloadable.download_content_type(),
                self.mime_type)

    def test_idownloadable_unicode(self):
        with self.admin_access.client_cnx() as cnx:
            ufobj = cnx.entity_from_eid(self.ufobj)
            idownloadable = ufobj.cw_adapt_to('IDownloadable')
            self.assertEqual(
                idownloadable.download_url(),
                u'http://testing.fr/cubicweb/%s/%s/raw/%s'
                % (ufobj.__regid__.lower(),
                   self.ufobj,
                   ufobj.data_name.replace(u'â', '%C3%A2')))

    def test_base(self):
        with self.admin_access.web_request() as req:
            fobj = req.entity_from_eid(self.fobj)
            self.assertEqual(fobj.size(), 3)
            self.assertEqual(
                fobj.icon_url(),
                'http://testing.fr/cubicweb/data/icons/'+self.icon)

    def test_views(self):
        with self.admin_access.web_request() as req:
            fobj = req.entity_from_eid(self.fobj)
            self.vreg['views'].select('download', req, rset=fobj.cw_rset)
            fobj.view('gallery')
            self.assertRaises(NoSelectableObject, fobj.view, 'image')
            self.assertRaises(NoSelectableObject, fobj.view, 'album')

    def test_hash(self):
        with self.admin_access.client_cnx() as cnx:
            self.create_user(cnx, login=u'simpleuser')
            cnx.commit()
        with self.new_access(u'simpleuser').client_cnx() as cnx:
            cnx.vreg.config['compute-hash'] = 1
            cnx.vreg.config['hash-algorithm'] = 'sha256'
            obj = cnx.create_entity('File', data_name=u"myfile.pdf",
                                    data=Binary(b"xxx"),
                                    data_format=self.mime_type)
            cnx.commit()
            self.assertEqual(
                '{sha256}cd2eb0837c9b4c962c22d2ff8b5441b7b45805887f051d39bf133b583baf6860',  # noqa
                obj.compute_hash())
            self.assertEqual(
                '{sha256}cd2eb0837c9b4c962c22d2ff8b5441b7b45805887f051d39bf133b583baf6860',  # noqa
                obj.data_hash)  # can read
            with self.assertRaises(Unauthorized):
                # write is forbiden
                obj.cw_set(data_hash=u'1234')
                cnx.commit()
            obj.cw_set(data=Binary(b'zzz'))
            obj.cw_clear_all_caches()
            self.assertEqual(
                '{sha256}17f165d5a5ba695f27c023a83aa2b3463e23810e360b7517127e90161eebabda',  # noqa
                obj.data_hash)
            with self.assertRaises(Unauthorized):
                cnx.create_entity('File', data_name=u'anotherfile.pdf',
                                  data=Binary(b'yyy'),
                                  data_format=self.mime_type,
                                  data_hash=u'deadbeef')
                cnx.commit()

    def test_hash_nodata(self):
        with self.admin_access.client_cnx() as cnx:
            with cnx.deny_all_hooks_but('metadata'):
                cnx.vreg.config['compute-hash'] = 1
                obj = cnx.create_entity('File')
                cnx.commit()
            self.assertEqual(None, obj.data)
            self.assertEqual(None, obj.data_hash)
            self.assertEqual(None, obj.compute_hash())

    def test_checkhash(self):
        with self.admin_access.client_cnx() as cnx:
            self.create_user(cnx, login=u'simpleuser')
            cnx.commit()
        with self.new_access(u'simpleuser').client_cnx() as cnx:
            cnx.vreg.config['compute-hash'] = 1
            cnx.vreg.config['hash-algorithm'] = 'md5'
            obj = cnx.create_entity('File', data_name=u"myfile.pdf",
                                    data=Binary(b"xxx"),
                                    data_format=self.mime_type)
            cnx.commit()
            self.assertEqual('{md5}f561aaf6ef0bf14d4208bb46a4ccb3ad',
                             obj.compute_hash())
            self.assertTrue(obj.check_hash())

            obj.cw_clear_all_caches()
            cnx.vreg.config['hash-algorithm'] = 'sha1'
            self.assertTrue(obj.check_hash())

            with cnx.allow_all_hooks_but('hash'):
                obj.cw_set(data=Binary(b'toto'))
                cnx.commit()

            # stored hash should still be the old one ('xxx' using md5)
            self.assertEqual('{md5}f561aaf6ef0bf14d4208bb46a4ccb3ad',
                             obj.data_hash)
            self.assertEqual('{sha1}0b9c2625dc21ef05f6ad4ddf47c5f203837aa32c',
                             obj.compute_hash(alg='sha1'))
            self.assertFalse(obj.check_hash())

    def test_checkhash_bwcompat(self):
        with self.admin_access.client_cnx() as cnx:
            self.create_user(cnx, login=u'simpleuser')
            cnx.commit()
        with self.new_access(u'simpleuser').client_cnx() as cnx:
            cnx.vreg.config['compute-hash'] = 1
            cnx.vreg.config['hash-algorithm'] = 'sha1'
            obj = cnx.create_entity('File', data_name=u"myfile.pdf",
                                    data=Binary(b"xxx"),
                                    data_format=self.mime_type)
            cnx.commit()
            self.assertEqual('{sha1}b60d121b438a380c343d5ec3c2037564b82ffef3',
                             obj.compute_hash())
            self.assertTrue(obj.check_hash())

            cnx.system_sql(
                'UPDATE cw_file SET cw_data_hash=%(hash)s '
                'WHERE cw_data_name=%(name)s',
                {'hash': 'b60d121b438a380c343d5ec3c2037564b82ffef3',
                 'name': 'myfile.pdf'})
            obj.cw_clear_all_caches()
            # note the absence of the hash algo "header" here:
            self.assertEqual('b60d121b438a380c343d5ec3c2037564b82ffef3',
                             obj.data_hash)
            self.assertTrue(obj.check_hash())


class ImageTC(CubicWebTC):
    icon = 'image_png.ico'
    mime_type = u"image/png"

    @property
    def data(self):
        with open(join(dirname(__file__), 'data', '20x20.gif'), 'rb') as fobj:
            return fobj.read()

    def setUp(self):
        super(ImageTC, self).setUp()
        cachedir = thumb_cache_dir(self.repo.vreg.config)
        if exists(cachedir):
            shutil.rmtree(cachedir)
        self.cachedir = cachedir

    def tearDown(self):
        super(ImageTC, self).tearDown()
        if exists(self.cachedir):
            shutil.rmtree(self.cachedir)

    def setup_database(self):
        with self.admin_access.repo_cnx() as cnx:
            create = cnx.create_entity
            self.fobj = create(
                'File', data_name=u"foo.gif", data=Binary(b"xxx"),
                data_format=self.mime_type).eid
            self.ufobj = create(
                'File', data_name=u"Bâbâr.png",
                data=Binary(b"yyy"),
                data_format=self.mime_type).eid
            cnx.commit()

    def test_views(self):
        with self.admin_access.web_request() as req:
            fobj = req.entity_from_eid(self.fobj)
            self.vreg['views'].select('download', req, rset=fobj.cw_rset)
            fobj.view('gallery')
            fobj.view('image')
            fobj.view('album')
        with self.admin_access.web_request(selected='something stupid') as req:
            fobj = req.entity_from_eid(self.fobj)
            with self.assertRaises(NotFound):
                req.view('gallery', fobj.cw_rset)
        with self.admin_access.web_request(selected='6666') as req:
            fobj = req.entity_from_eid(self.fobj)
            with self.assertRaises(NotFound):
                req.view('gallery', fobj.cw_rset)

    def test_thumbnail_generation_fails(self):
        with self.admin_access.web_request() as req:
            fobj = req.entity_from_eid(self.fobj)
            ithumb = fobj.cw_adapt_to('IThumbnail')
            # the actual thumbnail generation fails because the actual
            # file content is (literally) "xxx"
            self.assertEqual(ithumb.thumbnail_data(), '')
            self.assertEqual(
                u'http://testing.fr/cubicweb/%s/%s/thumb/foo_75x75.png' %
                (fobj.__regid__.lower(), self.fobj),
                ithumb.thumbnail_url())

    def test_thumbnail(self):
        with self.admin_access.client_cnx() as cnx:
            img = cnx.create_entity(
                'File', data=Binary(self.data),
                data_name=u'20x20.gif')
            cnx.commit()
            thumbadapter = img.cw_adapt_to('IThumbnail')
            self.assertEqual(
                '20x20_75x75.png',
                thumbadapter.thumbnail_file_name())
            self.assertEqual(
                'http://testing.fr/cubicweb/file/%s/thumb/20x20_75x75.png' % img.eid,  # noqa
                thumbadapter.thumbnail_url())
            cachepath = thumbadapter._thumbnail_path()
            self.assertIsNone(thumbadapter.thumbnail_path())
            self.assertFalse(isfile(cachepath))

            img = cnx.execute(
                'File F WHERE F data_name="20x20.gif"').get_entity(0, 0)
            thumbadapter = img.cw_adapt_to('IThumbnail')
            self.assertTrue(thumbadapter.thumbnail_data())
            cachepath = thumbadapter.thumbnail_path()
            self.assertTrue(isfile(cachepath))
            self.assertEqual(open(cachepath, 'rb').read(),
                             thumbadapter.thumbnail_data())


class MimeTypeDetectionTC(CubicWebTC):

    def test_extra_dot(self):
        with self.admin_access.client_cnx() as cnx:
            fobj = cnx.create_entity('File', data_name=u"foo.toto.pdf",
                                     data=Binary(b"xxx"))
            self.assertEqual(fobj.data_format, 'application/pdf')

    def test_file_name_priority(self):
        with self.admin_access.web_request() as req:
            req.form = {
                'eid': ['X'],
                '__maineid': 'X',
                '__type:X': 'File',
                '_cw_entity_fields:X': 'data-subject,data_name-subject',
                'data-subject:X': (u'coucou.txt', Binary(b'coucou')),
                'data_name-subject:X': u'coco.txt',
            }
            path, params = self.expect_redirect_handle_request(req, 'edit')
            self.assertTrue(path.startswith('file/'), path)
            eid = path.split('/')[1]
            efile = req.entity_from_eid(eid)
            self.assertEqual(efile.data_name, 'coco.txt')


if __name__ == '__main__':
    from unittest import main
    main()



# pylint: disable-msg=W0622
"""cubicweb-file packaging information"""

modname = 'cubicweb_file'
distname = "cubicweb-file"

numversion = (2, 2, 0)
version = '.'.join(str(num) for num in numversion)

license = 'LGPL'
author = "Logilab"
author_email = "contact@logilab.fr"
web = 'https://www.cubicweb.org/project/%s' % distname
description = "file component for the CubicWeb framework"
classifiers = [
           'Environment :: Web Environment',
           'Framework :: CubicWeb',
           'Programming Language :: Python',
           'Programming Language :: JavaScript',
]

__depends__ = {'cubicweb': '>= 3.26',
               'Pillow': None,
               }

__recommends__ = {}



"""entity classes for File entities

:organization: Logilab
:copyright: 2003-2018 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
:contact: http://www.logilab.fr/ -- mailto:contact@logilab.fr
"""

from warnings import warn
import tempfile
import os
from os.path import join, splitext, exists, isfile, isabs

from six import text_type
from PIL.Image import open as pilopen
from PIL.Image import ANTIALIAS

from logilab.mtconverter import guess_mimetype_and_encoding
from logilab.common.deprecation import deprecated

from cubicweb import Binary
from cubicweb.entities import AnyEntity, fetch_config, adapters
from cubicweb.view import EntityAdapter
from cubicweb.predicates import is_instance, has_mimetype

from cubicweb_file.hashtools import compute_hash, check_hash


class File(AnyEntity):
    """customized class for File entities"""
    __regid__ = 'File'
    fetch_attrs, cw_fetch_order = fetch_config(['data_name', 'title'])

    def set_format_and_encoding(self):
        """try to set format and encoding according to known values (filename,
        file content, format, encoding).

        This method must be called in a before_[add|update]_entity hook else it
        won't have any effect.
        """
        assert 'data' in self.cw_edited, "missing mandatory attribute data"
        if self.cw_edited.get('data'):
            if (hasattr(self.data, 'filename')
                    and not self.cw_edited.get('data_name')):
                self.cw_edited['data_name'] = self.data.filename
        else:
            self.cw_edited['data_format'] = None
            self.cw_edited['data_encoding'] = None
            self.cw_edited['data_name'] = None
            return
        if 'data_format' in self.cw_edited:
            format = self.cw_edited.get('data_format')
        else:
            format = None
        if 'data_encoding' in self.cw_edited:
            encoding = self.cw_edited.get('data_encoding')
        else:
            encoding = None
        if not (format and encoding):
            format, encoding = guess_mimetype_and_encoding(
                data=self.cw_edited.get('data'),
                # use get and not get_value since data has changed, we only
                # want to consider explicitly specified values, not old ones
                filename=self.cw_edited.get('data_name'),
                format=format, encoding=encoding,
                fallbackencoding=self._cw.encoding)
            if format:
                self.cw_edited['data_format'] = text_type(format)
            if encoding:
                self.cw_edited['data_encoding'] = text_type(encoding)

    def dc_title(self):
        if self.title:
            return '%s (%s)' % (self.title, self.data_name)
        return self.data_name

    def size(self):
        rql = "Any LENGTH(D) WHERE X eid %(x)s, X data D"
        return self._cw.execute(rql, {'x': self.eid})[0][0]

    def read(self, size=-1):
        return self.data.read(size)

    def seek(self, pos, *args, **kwargs):
        return self.data.seek(pos, *args, **kwargs)

    def tell(self):
        return self.data.tell()

    def icon_url(self):
        config = self._cw.vreg.config
        for rid in (self.data_format.replace('/', '_', 1),
                    self.data_format.split('/', 1)[0],
                    'default'):
            iconfile = rid + '.ico'
            rpath, iconfile = config.locate_resource(join('icons', iconfile))
            if rpath is not None:
                return self._cw.data_url(iconfile)

    def check_hash(self):
        """Return True if stored hash value is equal to the computed one from
        currently stored data.

        True if no hash is stored (typically when 'compute-hash' config option
        is false)

        """
        if self.data_hash:
            return check_hash(self.data_hash, self.data
                              and self.data.getvalue())
        return True

    def compute_hash(self, value=None, alg=None):

        """Compute the hash "value" data using "alg" hash algorithm.
        Returns a string "{ALG}HASH"

        If value is None, use stored data value.

        If alg is None, use currently configured hash algorithm (hash-algorithm
        in instance config options).

        """

        if value is None and self.data is not None:
            value = self.data.getvalue()
        if value is not None:
            if alg is None:
                alg = self._cw.vreg.config['hash-algorithm']
            return compute_hash(value, alg)

    @property
    def hash_value(self):
        """Return the hash of the currently stored data, as a string like "{ALG}HASH"

        If the hash as already been computed and stored as data_hash attribute,
        it is returned, otherwise, it is computed.

        Note that no check is performed to unsure returned hash is still valid
        for currently stored data.

        """
        if self.data_hash:
            return self.data_hash
        return self.compute_hash()

    @deprecated('[file 2.1] use entity.compute_hash() instead')
    def compute_sha1hex(self, value=None):
        return self.compute_hash(value, alg='sha1')

    @deprecated('[file 1.9] use entity.cw_adapt_to("IDownloadable")'
                '.download_url() instead')
    def download_url(self):
        return self.cw_adapt_to('IDownloadable').download_url()

    @deprecated('[file 1.9] use entity.cw_adapt_to("IDownloadable")'
                '.download_content_type() instead')
    def download_content_type(self):
        return self.cw_adapt_to('IDownloadable').download_content_type()

    @deprecated('[file 1.9] use entity.cw_adapt_to("IDownloadable")'
                '.download_encoding() instead')
    def download_encoding(self):
        return self.cw_adapt_to('IDownloadable').download_encoding()

    @deprecated('[file 1.9] use entity.cw_adapt_to("IDownloadable")'
                '.download_file_name() instead')
    def download_file_name(self):
        return self.cw_adapt_to('IDownloadable').download_file_name()

    @deprecated('[file 1.9] use entity.cw_adapt_to("IDownloadable")'
                '.download_data() instead')
    def download_data(self):
        return self.cw_adapt_to('IDownloadable').download_data()

    @property
    @deprecated('[file 1.6] use entity.data_name instead')
    def name(self):
        return self.data_name


class FileIDownloadableAdapter(adapters.IDownloadableAdapter):
    __select__ = is_instance('File')

    # IDownloadable
    def download_url(self, **kwargs):
        # include filename in download url for nicer url
        if 'small' in kwargs:
            warn('[1.15.0] `small` keyword is deprecated, you should use '
                 'FileThumbnailAdapter.download_thumbnail instead',
                 DeprecationWarning)
        name = self._cw.url_quote(self.download_file_name())
        path = '%s/raw/%s' % (self.entity.rest_path(), name)
        return self._cw.build_url(path, **kwargs)

    def download_content_type(self):
        return self.entity.data_format

    def download_encoding(self):
        return self.entity.data_encoding

    def download_file_name(self):
        return self.entity.data_name

    def download_data(self):
        if (self.download_content_type().startswith('image/')
                and 'small' in getattr(self._cw, 'form', ())):
            # Session has no .form, have to consume force param to avoid
            # infinite recursion
            self._cw.form.pop('small')
            warn('[1.15.0] `small` keyword is deprecated, you should use '
                 'FileThumbnailAdapter.download_thumbnail instead',
                 DeprecationWarning)
            return self.entity.cw_adapt_to('IThumbnail').thumbnail_data()
        return self.entity.data.getvalue()


def thumb_cache_dir(config):
    """Return the directory to use for thumbnails cache

    This function creates the directory if necessary.
    Returns None if cannot cache"""
    directory = config['thumbnail-cache-directory']
    if directory is None:
        directory = join(config.appdatahome, 'cache', 'file', 'thumbnail')
    else:
        if not isabs(directory):
            config.warning('thumbnail-cache-directory must be '
                           'an absolute existing path')
            return None
    try:
        os.makedirs(directory)
    except Exception as e:
        if not isinstance(e, OSError) or e.errno != os.errno.EEXIST:
            config.warning('thumbnail-cache-directory does not exists '
                           'and cannot be created, '
                           'check your filesystem permissions')
            return
    return directory


# This adapter should not be considered as an official API to manage
# thumbnails for any kind of document/content. A proper and generic
# API might be proposed in the future
class FileThumbnailAdapter(EntityAdapter):
    __regid__ = 'IThumbnail'
    __select__ = is_instance('File') & has_mimetype('image/')

    def _thumbnail_path(self):
        """Default implementation of the path generation for a thumbnail of an image
        file based on the hash of the image.

        If hash is not stored (typically when 'compute-hash' config option is
        False), the hash is computed from the image data.

        This generates a path like:

        <thumbnail-cache-directory>/<hash>_<thum_size>.png

        Note that if the stored hash value is out of sync with stored data
        value, it may produce an incorrect path (may occur if modifications are
        done with the 'hash' hooks disabled).

        """
        cachedir = thumb_cache_dir(self._cw.vreg.config)
        if cachedir:
            hashhex = self.entity.hash_value
            size = self._cw.vreg.config['image-thumb-size']
            return join(cachedir, '%s_%s.png' % (hashhex, size))
        return None

    def thumbnail_file_name(self):
        name, _ext = splitext(self.entity.data_name)
        size = self._cw.vreg.config['image-thumb-size']
        return '%s_%s.png' % (name, size)

    def thumbnail_url(self):
        name = self._cw.url_quote(self.thumbnail_file_name())
        path = '%s/thumb/%s' % (self.entity.rest_path(), name)
        return self._cw.build_url(path)

    def thumbnail_data(self):
        thumbpath = self._thumbnail_path()
        # use cache if it exists
        if thumbpath is not None:
            if isfile(thumbpath):
                try:
                    return open(thumbpath, 'rb').read()
                except IOError:
                    self.warning('Failed to load cached thumbnail file %s',
                                 thumbpath)
        iimage = self.entity.cw_adapt_to('IImage')
        try:
            data = iimage.thumbnail().getvalue()
        except UnResizeable:
            self.info('Failed to generate thumbnail for %s', self.entity.eid)
            return ''
        # fill cache if possible
        if thumbpath is not None:
            try:
                with tempfile.NamedTemporaryFile(
                        delete=False,
                        prefix=thumb_cache_dir(self._cw.vreg.config)) as tmp:
                    tmp.write(data)
                try:
                    os.rename(tmp.name, thumbpath)
                except Exception:
                    os.unlink(tmp.name)
                    raise
            except EnvironmentError:
                self.warning('Failed to save thumbnail in %s', thumbpath)
        return data

    def thumbnail_path(self):
        path = self._thumbnail_path()
        if path and exists(path):
            return path
        return None


class UnResizeable(Exception):
    pass


class IImageAdapter(EntityAdapter):
    __regid__ = 'IImage'
    __select__ = has_mimetype('image/')

    def __init__(self, *args, **kwargs):
        super(IImageAdapter, self).__init__(*args, **kwargs)
        idownloadable = self.entity.cw_adapt_to('IDownloadable')
        for meth in ('download_url', 'download_content_type',
                     'download_encoding', 'download_file_name',
                     'download_data'):
            setattr(self, meth, getattr(idownloadable, meth))

    def resize(self, size):
        size = tuple(int(s) for s in size.split('x'))
        idownloadable = self.entity.cw_adapt_to('IDownloadable')
        ctype = idownloadable.download_content_type()
        fmt = ctype and ctype.split('/', 1)[1] or None
        if fmt is None:
            self.error('unable to resize')
            raise UnResizeable
        data = idownloadable.download_data()
        pilimg = pilopen(Binary(data))
        pilimg.thumbnail(size, ANTIALIAS)
        stream = Binary()
        pilimg.save(stream, fmt)
        stream.seek(0)
        stream.filename = idownloadable.download_file_name()
        return stream

    def thumbnail(self, shadow=False, size=None):
        if size is None:
            size = self._cw.vreg.config['image-thumb-size']
        size = tuple(int(s) for s in size.split('x'))
        idownloadable = self.entity.cw_adapt_to('IDownloadable')
        data = idownloadable.download_data()
        try:
            pilimg = pilopen(Binary(data))
        except IOError:
            raise UnResizeable
        if shadow:
            self.warning('[1.15.0] the shadow parameter is now unused '
                         'and you should use css rules to lay shadows out',
                         DeprecationWarning)
        pilimg.thumbnail(size, ANTIALIAS)
        stream = Binary()
        pilimg.save(stream, 'png')
        stream.seek(0)
        ithumbnail = self.entity.cw_adapt_to('IThumbnail')
        stream.filename = ithumbnail.thumbnail_file_name()
        return stream



add_attribute('File', 'data_sha1hex')

if (config.get('compute-sha1hex') and
    confirm('compute sha1sum for each File ?', default='n')):
    for entity, data in rql('Any X,D WHERE X is File, X data D').iter_rows_with_entities():
        entity.cw_set(data_sha1hex=entity.compute_sha1hex(data.getvalue()))
        entity.cw_clear_all_caches()
    commit()




subclasses = False
is_rel = 'is'

if confirm('Also handle entity types that inherit from File?'):
    subclasses = True
    is_rel = 'is_instance_of'

add_attribute('File', 'data_hash')
if subclasses:
    for etype in schema['File'].specialized_by(recursive=True):
        add_attribute(etype.type, 'data_hash')

rql('SET X data_hash "{sha1}"+H WHERE X %(rel)s File, '
    'X data_sha1hex H, '
    'NOT X data_sha1hex NULL' % {'rel': is_rel})

drop_attribute('File', 'data_sha1hex')
if subclasses:
    for etype in schema['File'].specialized_by(recursive=True):
        drop_attribute(etype.type, 'data_sha1hex')

commit()



from cubicweb.view import EntityView
from cubicweb.predicates import is_instance, one_line_rset, adaptable
from cubicweb.web import formwidgets as wdgs
from cubicweb.web.views import baseviews, uicfg, idownloadable

_pvs = uicfg.primaryview_section
_pvs.tag_attribute(('File', 'title'), 'hidden')
_pvs.tag_attribute(('File', 'data_name'), 'hidden')
_pvs.tag_attribute(('File', 'data_sha1hex'), 'hidden')

_pvdc = uicfg.primaryview_display_ctrl
_pvdc.tag_attribute(('File', 'description'), {'showlabel': False})

# fields required in the schema but automatically set by hooks. Tell about that
# to the ui
_pvdc = uicfg.autoform_field_kwargs
_pvdc.tag_attribute(('File', 'data_name'), {
    'required': False, 'widget': wdgs.TextInput({'size': 45})})
_pvdc.tag_attribute(('File', 'data_format'), {'required': False})


class FileOneLine(baseviews.OneLineView):
    __select__ = is_instance('File')

    def cell_call(self, row, col):
        entity = self.cw_rset.get_entity(row, col)
        self.w(u'<img src="{}" alt=""/>'.format(entity.icon_url()))
        super(FileOneLine, self).cell_call(row, col)


class FileInContext(baseviews.InContextView):
    __select__ = is_instance('File')

    def cell_call(self, row, col):
        entity = self.cw_rset.get_entity(row, col)
        self.w(u'<img src="{}" alt=""/>'.format(entity.icon_url()))
        super(FileInContext, self).cell_call(row, col)


class FileOutOfContext(baseviews.OutOfContextView):
    __select__ = is_instance('File')

    def cell_call(self, row, col):
        entity = self.cw_rset.get_entity(row, col)
        self.w(u'<img src="{}" alt=""/>'.format(entity.icon_url()))
        super(FileOutOfContext, self).cell_call(row, col)


class FileIcon(EntityView):
    __select__ = is_instance('File')
    __regid__ = 'icon'

    def cell_call(self, row, col):
        entity = self.cw_rset.get_entity(row, col)
        self.w(u'<a href="{}" title="{}"><img src="{}" alt=""/></a>'.format(
            entity.absolute_url(), entity.dc_title(), entity.icon_url()))


class ThumbnailView(idownloadable.DownloadView):
    """thumbnail download view """
    __regid__ = 'thumbnail'
    __select__ = one_line_rset() & adaptable('IThumbnail')

    content_type = 'image/png'

    def set_request_content_type(self):
        """overriden to set the correct filetype and filename"""
        entity = self.cw_rset.complete_entity(self.cw_row or 0, self.cw_col or 0)
        adapter = entity.cw_adapt_to('IThumbnail')
        self._cw.set_content_type(self.content_type,
                                  filename=adapter.thumbnail_file_name(),
                                  disposition='attachment')

    def call(self):
        entity = self.cw_rset.complete_entity(self.cw_row or 0, self.cw_col or 0)
        adapter = entity.cw_adapt_to('IThumbnail')
        self.w(adapter.thumbnail_data())
